import{d as w,cL as m,ao as C,c as M,k as a,l as u,o as y,m as l,n as o,Z as r,cM as h,a6 as i,cN as x,cO as R,cP as G}from"./index-CgvuCQfr.js";const T=w({__name:"ConcurrencyLimits",setup(g){const{showModal:t,open:p}=m(),{showModal:e,open:b}=m(),_=[{label:"Global"},{label:"Task Run"}],c=C("tab","Global");return(v,s)=>{const d=u("p-button"),k=u("p-tabs"),f=u("p-layout-default");return y(),M(f,{class:"concurrency-limits"},{header:a(()=>[l(o(r),{crumbs:[{text:"Concurrency"}]})]),default:a(()=>[l(k,{selected:o(c),"onUpdate:selected":s[2]||(s[2]=n=>i(c)?c.value=n:null),tabs:_},{global:a(()=>[l(o(r),{size:"lg",crumbs:[{text:"Global Concurrency Limits"}]},{"after-crumbs":a(()=>[l(d,{small:"",icon:"PlusIcon",onClick:o(p)},null,8,["onClick"])]),_:1}),l(o(h),{showModal:o(t),"onUpdate:showModal":s[0]||(s[0]=n=>i(t)?t.value=n:null)},null,8,["showModal"]),l(o(x),{class:"concurrency-limits__global-table"})]),"task-run":a(()=>[l(o(r),{size:"lg",crumbs:[{text:"Task Run Concurrency Limits"}]},{"after-crumbs":a(()=>[l(d,{small:"",icon:"PlusIcon",onClick:o(b)},null,8,["onClick"])]),_:1}),l(o(R),{showModal:o(e),"onUpdate:showModal":s[1]||(s[1]=n=>i(e)?e.value=n:null)},null,8,["showModal"]),l(o(G),{class:"concurrency-limits__task-limits-table"})]),_:1},8,["selected"])]),_:1})}}});export{T as default};
//# sourceMappingURL=ConcurrencyLimits-sBs6g7bZ.js.map
