import{d as i,g as b,f as d,j as o,bK as _,c as f,k as c,a as v,l as D,i as B,o as w,m as n,n as s,cw as g,cx as h,bo as x}from"./index-CgvuCQfr.js";import{u as C}from"./usePageTitle-DVT0B9Gy.js";const R=i({__name:"BlockView",setup(y){const u=b(),l=B(),t=d("blockDocumentId"),a=o(()=>t.value?[t.value]:null),r=_(u.blockDocuments.getBlockDocument,a),e=o(()=>r.response),m=()=>{l.push(x.blocks())},k=o(()=>e.value?`Block: ${e.value.name}`:"Block");return C(k),(V,I)=>{const p=D("p-layout-default");return e.value?(w(),f(p,{key:0,class:"block-view"},{header:c(()=>[n(s(g),{"block-document":e.value,onDelete:m},null,8,["block-document"])]),default:c(()=>[n(s(h),{"block-document":e.value},null,8,["block-document"])]),_:1})):v("",!0)}}});export{R as default};
//# sourceMappingURL=BlockView-Cfhmmb-4.js.map
