import{d as m,g as p,t as d,j as a,c as s,k as c,l as k,o as t,m as _,n as o,cj as f,A as i,ck as y,cl as B,F as x,a as C}from"./index-CgvuCQfr.js";import{u as D}from"./usePageTitle-DVT0B9Gy.js";const V=m({__name:"Blocks",setup(g){const n=p(),e=d(n.blockDocuments.getBlockDocumentsCount),l=a(()=>e.executed&&e.response==0),r=a(()=>e.executed);return D("Blocks"),(h,b)=>{const u=k("p-layout-default");return t(),s(u,{class:"blocks"},{header:c(()=>[_(o(f))]),default:c(()=>[r.value?(t(),i(x,{key:0},[l.value?(t(),s(o(y),{key:0})):(t(),s(o(B),{key:1,onDelete:o(e).refresh},null,8,["onDelete"]))],64)):C("",!0)]),_:1})}}});export{V as default};
//# sourceMappingURL=Blocks-rL2qEeun.js.map
