import{d as u,s,w as c,c as l,r as m,a as p,o as i,e as f,b as v,u as d,W as h}from"./index-CgvuCQfr.js";const r=a=>{const{value:e,isCustom:o}=f();if(v(a.query)&&o.value)try{const t=d.map("SavedSearchFilter",e.value,"LocationQuery");return{...a,query:t}}catch(t){console.error(t)}return!0},y=u({expose:[],beforeRouteEnter:r,beforeRouteUpdate:r}),b=u({...y,__name:"RunsPageWithDefaultFilter",props:{component:{type:Function}},setup(a){const e=a,o=s(null);function t(n){return h(n)}return c(e.component,()=>{t(e.component)?e.component().then(n=>{o.value=n.default}):o.value=e.component},{immediate:!0}),(n,C)=>o.value!==null?(i(),l(m(o.value),{key:0})):p("",!0)}});export{b as default};
//# sourceMappingURL=RunsPageWithDefaultFilter-CE6TV8Ff-oT-MPbGd.js.map
