import{d as k,g as m,af as f,j as o,cm as y,t as _,c as d,k as t,l as g,o as v,m as s,n as l,cn as C,co as B}from"./index-CgvuCQfr.js";import{u as T}from"./usePageTitle-DVT0B9Gy.js";const S=k({__name:"BlocksCatalog",setup(x){const c=m(),a=f(null),n=o(()=>a.value?[a.value]:[]),{filter:p}=y({blockSchemas:{blockCapabilities:n}}),i=_(c.blockTypes.getBlockTypes,[p]),u=o(()=>i.response??[]);return T("Blocks Catalog"),(h,e)=>{const r=g("p-layout-default");return v(),d(r,{class:"blocks-catalog"},{header:t(()=>[s(l(C))]),default:t(()=>[s(l(B),{capability:a.value,"onUpdate:capability":e[0]||(e[0]=b=>a.value=b),"block-types":u.value},null,8,["capability","block-types"])]),_:1})}}});export{S as default};
//# sourceMappingURL=BlocksCatalog-sNofwook.js.map
