import{j as o,bA as a,n as c}from"./index-CgvuCQfr.js";function i(...t){const r=[...t,"Prefect Server"],n=o(()=>r.map(e=>c(e)).filter(e=>e!==null).join(" • "));a(()=>document.title=n.value)}export{i as u};
//# sourceMappingURL=usePageTitle-DVT0B9Gy.js.map
