import{d as g,g as v,t as k,j as s,c as n,k as i,l as r,o as a,m as l,n as t,cA as N,D as x,A as y,cB as C,cC as B,F as h,a as w}from"./index-CgvuCQfr.js";import{u as A}from"./usePageTitle-DVT0B9Gy.js";const F=g({__name:"Notifications",setup(P){const u=v(),e=k(u.notifications.getNotifications),c=s(()=>e.response??[]),f=s(()=>e.executed&&c.value.length===0),d=s(()=>e.executed);return A("Notifications"),(V,o)=>{const p=r("p-message"),m=r("p-layout-default");return a(),n(m,{class:"notifications"},{header:i(()=>[l(t(N))]),default:i(()=>[l(p,{info:""},{default:i(()=>[x(" Notifications are deprecated and will be migrated in the future. Please use Automations. ")]),_:1}),d.value?(a(),y(h,{key:0},[f.value?(a(),n(t(C),{key:0})):(a(),n(t(B),{key:1,notifications:c.value,onDelete:o[0]||(o[0]=_=>t(e).refresh()),onUpdate:o[1]||(o[1]=_=>t(e).refresh())},null,8,["notifications"]))],64)):w("",!0)]),_:1})}}});export{F as default};
//# sourceMappingURL=Notifications-DtGolR5s.js.map
