import{d as S,g as B,f as D,t as R,j as p,af as T,ao as V,c,k as e,l as n,I as _,aK as $,o as l,n as t,aL as h,a as i,aM as d,A as K,m as s,aN as M,aO as P,aP as v,D as Q,y as j,aQ as m,a6 as z}from"./index-CgvuCQfr.js";import{u as E}from"./usePageTitle-DVT0B9Gy.js";const H={key:0},q=S({__name:"Artifact",setup(L){const b=B(),k=D("artifactId"),w=R(b.artifacts.getArtifact,[k]),a=p(()=>w.response),o=T(!1),y=[{label:"Artifact"},{label:"Details"},{label:"Raw"}],u=V("tab","Artifact"),C=p(()=>a.value?`${_.info.artifact}: ${a.value.key??$(a.value.type)}`:_.info.artifact);return E(C),(O,r)=>{const g=n("p-divider"),x=n("p-button"),A=n("p-content"),I=n("p-tabs"),N=n("p-layout-well");return l(),c(N,{class:"artifact"},{header:e(()=>[a.value?(l(),c(t(h),{key:0,artifact:a.value},null,8,["artifact"])):i("",!0)]),well:e(()=>[a.value?(l(),c(t(d),{key:0,artifact:a.value,alternate:""},null,8,["artifact"])):i("",!0)]),default:e(()=>[a.value?(l(),K("section",H,[s(t(M),{artifact:a.value},null,8,["artifact"]),s(g),t(P).xl?(l(),c(A,{key:0},{default:e(()=>[s(t(v),{artifact:a.value},null,8,["artifact"]),s(x,{class:"artifact__raw-data-button",small:"",onClick:r[0]||(r[0]=f=>o.value=!o.value)},{default:e(()=>[Q(j(o.value?"Hide":"Show")+" raw data ",1)]),_:1}),o.value?(l(),c(t(m),{key:0,artifact:a.value},null,8,["artifact"])):i("",!0)]),_:1})):(l(),c(I,{key:1,selected:t(u),"onUpdate:selected":r[1]||(r[1]=f=>z(u)?u.value=f:null),tabs:y},{artifact:e(()=>[s(t(v),{artifact:a.value},null,8,["artifact"])]),details:e(()=>[s(t(d),{artifact:a.value},null,8,["artifact"])]),raw:e(()=>[s(t(m),{artifact:a.value},null,8,["artifact"])]),_:1},8,["selected"]))])):i("",!0)]),_:1})}}});export{q as default};
//# sourceMappingURL=Artifact-CiW0TeFj.js.map
