import{d as m,g as y,f as k,t as _,j as o,c as e,k as s,l as v,I as n,o as t,n as c,aR as d,a as r,aN as g,A as C,aS as h,F as x}from"./index-CgvuCQfr.js";import{u as A}from"./usePageTitle-DVT0B9Gy.js";const R=m({__name:"ArtifactKey",setup(B){const l=y(),u=k("artifactKey"),i=_(l.artifacts.getArtifactCollection,[u]),a=o(()=>i.response),f=o(()=>a.value?`${n.info.artifact}: ${a.value.key}`:n.info.artifact);return A(f),(K,w)=>{const p=v("p-layout-default");return t(),e(p,{class:"artifact"},{header:s(()=>[a.value?(t(),e(c(d),{key:0,artifact:a.value},null,8,["artifact"])):r("",!0)]),default:s(()=>[a.value?(t(),e(c(g),{key:0,artifact:a.value},null,8,["artifact"])):r("",!0),a.value?(t(),C(x,{key:1},[a.value.key?(t(),e(c(h),{key:0,"artifact-key":a.value.key},null,8,["artifact-key"])):r("",!0)],64)):r("",!0)]),_:1})}}});export{R as default};
//# sourceMappingURL=ArtifactKey-hYNm1Xy_.js.map
