import{d,g as y,t as _,j as o,c as a,k as n,l as i,o as t,m as f,n as s,bZ as k,A as g,b_ as b,bY as h,F as v,a as x}from"./index-CgvuCQfr.js";import{u as D}from"./usePageTitle-DVT0B9Gy.js";const V=d({__name:"Deployments",setup(B){const l=y(),c={interval:3e4},e=_(l.deployments.getDeployments,[{}],c),p=o(()=>e.response??[]),r=o(()=>e.executed&&p.value.length===0),m=o(()=>e.executed);return D("Deployments"),(C,w)=>{const u=i("p-layout-default");return t(),a(u,{class:"deployments"},{header:n(()=>[f(s(k))]),default:n(()=>[m.value?(t(),g(v,{key:0},[r.value?(t(),a(s(b),{key:0})):(t(),a(s(h),{key:1,onDelete:s(e).refresh},null,8,["onDelete"]))],64)):x("",!0)]),_:1})}}});export{V as default};
//# sourceMappingURL=Deployments-DVEao9L-.js.map
