import{d as v,j as h,o as i,A as u,m as e,k as t,F as C,B as x,n,da as B,y as f,M as w,c as A,l,Z as M,db as P,dc as d,a6 as V}from"./index-CgvuCQfr.js";import{u as j}from"./usePageTitle-DVT0B9Gy.js";import{u as F}from"./usePrefectApi-BpSnfZdy.js";const T={class:"settings-block"},Z=v({__name:"SettingsCodeBlock",props:{engineSettings:{}},setup(p){const s=p,a=h(()=>Object.entries(s.engineSettings));return(g,c)=>(i(),u("div",T,[e(n(B),{multiline:""},{default:t(()=>[(i(!0),u(C,null,x(a.value,(o,r)=>(i(),u("div",{key:r,class:"settings-block--code-line"},f(o[0])+": "+f(o[1]),1))),128))]),_:1})]))}}),N=v({__name:"Settings",async setup(p){let s,a;const g=[{text:"Settings"}],c=F(),[o,r]=([s,a]=w(()=>Promise.all([c.admin.getSettings(),c.admin.getVersion()])),s=await s,a(),s);return j("Settings"),(D,m)=>{const b=l("p-key-value"),S=l("p-theme-toggle"),_=l("p-label"),k=l("p-layout-default");return i(),A(k,{class:"settings"},{header:t(()=>[e(n(M),{crumbs:g},{actions:t(()=>[e(b,{class:"settings__version",label:"Version",value:n(r),alternate:""},null,8,["value"])]),_:1})]),default:t(()=>[e(_,{label:"Theme"},{default:t(()=>[e(S)]),_:1}),e(_,{label:"Color Mode",class:"settings__color-mode"},{default:t(()=>[e(n(P),{selected:n(d),"onUpdate:selected":m[0]||(m[0]=y=>V(d)?d.value=y:null)},null,8,["selected"])]),_:1}),e(_,{label:"Server Settings"},{default:t(()=>[e(Z,{class:"settings__code-block","engine-settings":n(o)},null,8,["engine-settings"])]),_:1})]),_:1})}}});export{N as default};
//# sourceMappingURL=Settings-x0w2N0q4.js.map
