import{d as m,g as _,t as k,j as s,c as a,k as n,l as i,o,m as f,n as t,cY as P,A as w,cZ as y,c_ as g,F as h,a as v}from"./index-CgvuCQfr.js";import{u as x}from"./usePageTitle-DVT0B9Gy.js";const F=m({__name:"WorkPools",setup(C){const r=_(),c={interval:3e4},e=k(r.workPools.getWorkPools,[{}],c),l=s(()=>e.response??[]),p=s(()=>e.executed&&l.value.length==0),u=s(()=>e.executed);return x("Work Pools"),(B,E)=>{const d=i("p-layout-default");return o(),a(d,{class:"work-pools"},{header:n(()=>[f(t(P))]),default:n(()=>[u.value?(o(),w(h,{key:0},[p.value?(o(),a(t(y),{key:0})):(o(),a(t(g),{key:1,onUpdate:t(e).refresh},null,8,["onUpdate"]))],64)):v("",!0)]),_:1})}}});export{F as default};
//# sourceMappingURL=WorkPools-Du2RYvzu.js.map
