import{d as l,g as p,c as _,k as o,l as m,o as d,m as e,n,cD as C,cE as h,cb as i,bo as c,N}from"./index-CgvuCQfr.js";import{u as b}from"./usePageTitle-DVT0B9Gy.js";const g=l({__name:"NotificationCreate",setup(w){const r=p();async function s(t){try{await r.notifications.createNotification(t),i.push(c.notifications())}catch(a){N("Error creating notification","error"),console.warn(a)}}function u(){i.push(c.notifications())}return b("Create Notification"),(t,a)=>{const f=m("p-layout-default");return d(),_(f,{class:"notification-create"},{header:o(()=>[e(n(C))]),default:o(()=>[e(n(h),{action:"Create",onSubmit:s,onCancel:u})]),_:1})}}});export{g as default};
//# sourceMappingURL=NotificationCreate-Eafbu0rB.js.map
