import{d as w,g,f as I,j as i,ao as R,bx as h,t as D,c,k as n,l as u,i as O,aO as S,o as a,n as t,cI as T,a as r,cJ as m,m as B,cK as N,a6 as V,bo as j}from"./index-CgvuCQfr.js";import{u as A}from"./usePageTitle-DVT0B9Gy.js";const P=w({__name:"ConcurrencyLimit",setup(H){const p=g(),y=I("concurrencyLimitId"),d=O(),v=i(()=>[{label:"Details",hidden:S.xl},{label:"Active Task Runs"}]),s=R("tab","Details"),{tabs:_}=h(v,s),b={interval:3e5},L=D(p.concurrencyLimits.getConcurrencyLimit,[y.value],b),e=i(()=>L.response);function f(){d.push(j.concurrencyLimits())}const k=i(()=>e.value?`Concurrency Limit: ${e.value.tag}`:"Concurrency Limit");return A(k),(J,l)=>{const C=u("p-tabs"),x=u("p-layout-well");return a(),c(x,{class:"concurrencyLimit"},{header:n(()=>[e.value?(a(),c(t(T),{key:0,"concurrency-limit":e.value,onDelete:f},null,8,["concurrency-limit"])):r("",!0)]),well:n(()=>[e.value?(a(),c(t(m),{key:0,alternate:"","concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),default:n(()=>[B(C,{selected:t(s),"onUpdate:selected":l[0]||(l[0]=o=>V(s)?s.value=o:null),tabs:t(_)},{details:n(()=>[e.value?(a(),c(t(m),{key:0,"concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),"active-task-runs":n(()=>{var o;return[(o=e.value)!=null&&o.activeSlots?(a(),c(t(N),{key:0,"active-slots":e.value.activeSlots},null,8,["active-slots"])):r("",!0)]}),_:1},8,["selected","tabs"])]),_:1})}}});export{P as default};
//# sourceMappingURL=ConcurrencyLimit-tt2mpyuj.js.map
