function s(n,e){if(e.length<n)throw new TypeError(n+" argument"+(n>1?"s":"")+" required, but only "+e.length+" present")}var o=1e3,r=3600,t=r*24;function a(n){return s(1,arguments),n*o}export{t as a,a as s};
//# sourceMappingURL=index-Vv9yEoRy.js.map
