import{d as k,g as c,f as t,M as w,c as _,k as r,l as d,o as i,m as n,n as o,d8 as P,d9 as f}from"./index-CgvuCQfr.js";import{u as Q}from"./usePageTitle-DVT0B9Gy.js";const x=k({__name:"WorkPoolQueueEdit",async setup(N){let e,u;const s=c(),a=t("workPoolName"),l=t("workPoolQueueName"),m=([e,u]=w(()=>s.workPoolQueues.getWorkPoolQueueByName(a.value,l.value)),e=await e,u(),e);return Q("Edit Work Pool Queue"),(y,h)=>{const p=d("p-layout-default");return i(),_(p,null,{header:r(()=>[n(o(P),{"work-pool-name":o(a),"work-pool-queue-name":o(l)},null,8,["work-pool-name","work-pool-queue-name"])]),default:r(()=>[n(o(f),{"work-pool-name":o(a),"work-pool-queue":o(m)},null,8,["work-pool-name","work-pool-queue"])]),_:1})}}});export{x as default};
//# sourceMappingURL=WorkPoolQueueEdit-DEMQn4NY.js.map
