import{d as e,c as o,n as a,o as t,dd as n}from"./index-CgvuCQfr.js";import{u as r}from"./usePageTitle-DVT0B9Gy.js";const m=e({__name:"404",setup(s){return r("Page Not Found"),(c,p)=>(t(),o(a(n)))}});export{m as default};
//# sourceMappingURL=404-DGN0CRxb.js.map
