import{d as m,af as d,g as _,j as o,I as v,c as r,k as c,l as b,t as k,o as s,m as h,n,cQ as y,A as C,cR as g,cS as x,F as V,a as B}from"./index-CgvuCQfr.js";import{u as P}from"./usePageTitle-DVT0B9Gy.js";const N=m({__name:"Variables",setup(S){const l=d(),a=()=>{var e;t.value.refresh(),(e=l.value)==null||e.refreshSubscriptions()},u=_(),t=o(()=>k(u.variables.getVariables)),p=o(()=>{var e;return t.value.executed&&((e=t.value.response)==null?void 0:e.length)===0}),i=o(()=>t.value.executed);return P(v.info.variables),(e,w)=>{const f=b("p-layout-default");return s(),r(f,{class:"variables"},{header:c(()=>[h(n(y),{onCreate:a})]),default:c(()=>[i.value?(s(),C(V,{key:0},[p.value?(s(),r(n(g),{key:0,onCreate:a})):(s(),r(n(x),{key:1,ref_key:"table",ref:l,onDelete:a,onUpdate:a},null,512))],64)):B("",!0)]),_:1})}}});export{N as default};
//# sourceMappingURL=Variables-BDMUgXUK.js.map
