import{d as f,g as i,t as _,j as o,c as s,k as n,l as w,o as t,m as k,n as a,bQ as C,A as b,bR as y,bS as h,F as v,a as x}from"./index-CgvuCQfr.js";import{u as F}from"./usePageTitle-DVT0B9Gy.js";const N=f({__name:"Flows",setup(g){const l=i(),c={interval:3e4},e=_(l.flows.getFlowsCount,[{}],c),r=o(()=>e.response??0),u=o(()=>e.executed&&r.value===0),p=o(()=>e.executed),m=()=>{e.refresh()};return F("Flows"),(B,R)=>{const d=w("p-layout-default");return t(),s(d,{class:"flows"},{header:n(()=>[k(a(C))]),default:n(()=>[p.value?(t(),b(v,{key:0},[u.value?(t(),s(a(y),{key:0})):(t(),s(a(h),{key:1,selectable:"",onDelete:m}))],64)):x("",!0)]),_:1})}}});export{N as default};
//# sourceMappingURL=Flows-C4cXy47o.js.map
