import{d as r,c as s,k as e,l as n,o as l,m as o,n as a,c$ as c,d0 as u}from"./index-CgvuCQfr.js";import{u as _}from"./usePageTitle-DVT0B9Gy.js";const k=r({__name:"WorkPoolCreate",setup(p){return _("Create Work Pool"),(f,m)=>{const t=n("p-layout-default");return l(),s(t,null,{header:e(()=>[o(a(c))]),default:e(()=>[o(a(u))]),_:1})}}});export{k as default};
//# sourceMappingURL=WorkPoolCreate-CsSmjRjX.js.map
