import{d as c,g as _,f as u,M as m,c as k,k as l,l as w,o as d,m as r,n as e,d1 as f,d2 as i}from"./index-CgvuCQfr.js";import{u as P}from"./usePageTitle-DVT0B9Gy.js";const W=c({__name:"WorkPoolEdit",async setup(y){let o,a;const s=_(),n=u("workPoolName"),t=([o,a]=m(()=>s.workPools.getWorkPoolByName(n.value)),o=await o,a(),o);return P("Create Work Pool"),(C,h)=>{const p=w("p-layout-default");return d(),k(p,null,{header:l(()=>[r(e(f),{"work-pool":e(t)},null,8,["work-pool"])]),default:l(()=>[r(e(i),{"work-pool":e(t)},null,8,["work-pool"])]),_:1})}}});export{W as default};
//# sourceMappingURL=WorkPoolEdit-CCFDvtaf.js.map
