.events__content{display:grid;grid-template-columns:repeat(1,minmax(0,1fr));gap:.25rem}.events__filters{margin-bottom:1.75rem;display:flex;flex-direction:column;gap:1rem}@media (min-width: 640px){.events__filters{flex-direction:row}}.events__chart{position:sticky;top:0;z-index:10;display:grid;grid-template-columns:repeat(1,minmax(0,1fr));gap:.75rem;border-top-left-radius:var(--p-radius-default);border-top-right-radius:var(--p-radius-default);padding-bottom:.75rem}.events__chart--stuck{border-top-left-radius:0;border-top-right-radius:0;border-bottom-right-radius:var(--p-radius-default);border-bottom-left-radius:var(--p-radius-default);--tw-shadow: 0 4px 6px -1px rgb(0 0 0 / .1), 0 2px 4px -2px rgb(0 0 0 / .1);--tw-shadow-colored: 0 4px 6px -1px var(--tw-shadow-color), 0 2px 4px -2px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow);--tw-backdrop-blur: blur(4px);-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);background-color:var(--p-color-bg-floating-sticky)}.events__controls{display:flex;justify-content:center;padding:.75rem}.events__timeline{margin-top:1.75rem}.events__pager{margin-top:1rem}
