import{d as f,g as x,h as b,f as c,j as r,aG as h,M as y,n as e,c as C,k as n,a as D,l as I,aH as g,o as k,m as s,aI as w,Z as E,aJ as B}from"./index-CgvuCQfr.js";import{u as F}from"./usePageTitle-DVT0B9Gy.js";const j=f({__name:"Event",async setup(L){let a,o;const l=x(),u=b(),v=c("eventDate"),_=c("eventId"),m=r(()=>g(v.value)),d=h({startDate:m,eventId:[_.value]}),t=([a,o]=y(()=>l.events.getFirstEvent(d.value)),a=await a,o(),a),p=r(()=>[{text:"Event Feed",to:u.events()},{text:t.eventLabel}]);return F(`Event: ${t.eventLabel}`),(N,V)=>{const i=I("p-layout-default");return e(t)?(k(),C(i,{key:0,class:"event"},{header:n(()=>[s(e(E),{crumbs:p.value},{actions:n(()=>[s(e(w),{event:e(t)},null,8,["event"])]),_:1},8,["crumbs"])]),default:n(()=>[s(e(B),{event:e(t)},null,8,["event"])]),_:1})):D("",!0)}}});export{j as default};
//# sourceMappingURL=Event-B1UirkAc.js.map
