import{d as k,g as y,f as i,j as o,bK as b,c as m,k as a,a as _,l as f,o as T,m as t,n as c,cp as d,cq as g}from"./index-CgvuCQfr.js";import{u as v}from"./usePageTitle-DVT0B9Gy.js";const x=k({__name:"BlocksCatalogView",setup(B){const s=y(),l=i("blockTypeSlug"),n=o(()=>l.value?[l.value]:null),u=b(s.blockTypes.getBlockTypeBySlug,n),e=o(()=>u.response),p=o(()=>e.value?`Block Type: ${e.value.name}`:null);return v(p),(C,S)=>{const r=f("p-layout-default");return e.value?(T(),m(r,{key:0,class:"blocks-catalog-view"},{header:a(()=>[t(c(d),{"block-type":e.value},null,8,["block-type"])]),default:a(()=>[t(c(g),{"block-type":e.value},null,8,["block-type"])]),_:1})):_("",!0)}}});export{x as default};
//# sourceMappingURL=BlocksCatalogView-Kj-IckMM.js.map
